#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


import os
import imp
import time
import copy
import hashlib
import inspect
import traceback
from multiprocessing import sharedctypes # numpy xfer

# gpi
import gpi
from gpi import QtCore, QtGui, QtWidgets
from .defines import ExternalNodeType, GPI_PROCESS, GPI_THREAD, stw, GPI_SHDM_PATH
from .defines import GPI_WIDGET_EVENT, REQUIRED, OPTIONAL, GPI_PORT_EVENT
from .dataproxy import DataProxy, ProxyType
from .logger import manager
from .port import InPort, OutPort
from .widgets import HidableGroupBox
from . import widgets as BUILTIN_WIDGETS
from . import syntax


# start logger for this module
log = manager.getLogger(__name__)

# for PROCESS data hack
import numpy as np

# Developer Interface Exceptions
class GPIError_nodeAPI_setData(Exception):
    def __init__(self, value):
        super(GPIError_nodeAPI_setData, self).__init__(value)

class GPIError_nodeAPI_getData(Exception):
    def __init__(self, value):
        super(GPIError_nodeAPI_getData, self).__init__(value)

class GPIError_nodeAPI_setAttr(Exception):
    def __init__(self, value):
        super(GPIError_nodeAPI_setAttr, self).__init__(value)

class GPIError_nodeAPI_getAttr(Exception):
    def __init__(self, value):
        super(GPIError_nodeAPI_getAttr, self).__init__(value)

class GPIError_nodeAPI_getVal(Exception):
    def __init__(self, value):
        super(GPIError_nodeAPI_getVal, self).__init__(value)

class NodeAPI(QtWidgets.QWidget):
    """
    Base class for all external nodes.

    External nodes implement the extensible methods of this class: i.e.
    compute(), validate(), execType(), etc... to create a unique Node module.
    """
    GPIExtNodeType = ExternalNodeType  # ensures the subclass is of THIS class
    modifyWdg = gpi.Signal(str, dict)

    def __init__(self, node):
        super(NodeAPI, self).__init__()

        #self.setToolTip("Double Click to Show/Hide each Widget")
        self.node = node

        self.label = ''
        self._detailLabel = ''
        self._docText = None
        self.parmList = []  # deprecated, since dicts have direct name lookup
        self.parmDict = {}  # mirror parmList for now
        self.parmSettings = {}  # for buffering wdg parms before copying to a PROCESS
        self.shdmDict = {} # for storing base addresses

        # grid for module widgets
        self.layout = QtWidgets.QGridLayout()

        # this must exist before user-widgets are added so that they can get
        # node label updates
        self.wdglabel = QtWidgets.QLineEdit(self.label)

        # allow logger to be used in initUI()
        self.log = manager.getLogger(node.getModuleName())

        try:
            self._initUI_ret = self.initUI()
        except:
            log.error('initUI() failed. '+str(node.item.fullpath)+'\n'+str(traceback.format_exc()))
            self._initUI_ret = -1  # error

        # make a label box with the unique id
        labelGroup = HidableGroupBox("Node Label")
        labelLayout = QtWidgets.QGridLayout()
        self.wdglabel.textChanged.connect(self.setLabel)
        labelLayout.addWidget(self.wdglabel, 0, 1)
        labelGroup.setLayout(labelLayout)
        self.layout.addWidget(labelGroup, len(self.parmList) + 1, 0)
        labelGroup.set_collapsed(True)
        labelGroup.setToolTip("Displays the Label on the Canvas (Double Click)")

        # make an about box with the unique id
        self.aboutGroup = HidableGroupBox("About")
        aboutLayout = QtWidgets.QGridLayout()
        self.about_button = QtWidgets.QPushButton("Open Node &Documentation")
        self.about_button.clicked.connect(self.openNodeDocumentation)
        aboutLayout.addWidget(self.about_button, 0, 1)
        self.aboutGroup.setLayout(aboutLayout)
        self.layout.addWidget(self.aboutGroup, len(self.parmList) + 2, 0)
        self.aboutGroup.set_collapsed(True)
        self.aboutGroup.setToolTip("Node Documentation (docstring + autodocs, Double Click)")

        # window (just a QTextEdit) that will show documentation text
        self.doc_text_win = QtWidgets.QTextEdit()
        self.doc_text_win.setPlainText(self.generateHelpText())
        self.doc_text_win.setReadOnly(True)
        doc_text_font = QtGui.QFont("Monospace", 14)
        self.doc_text_win.setFont(doc_text_font)
        self.doc_text_win.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.doc_text_win.setWindowTitle(node.getModuleName() + " Documentation")

        hbox = QtWidgets.QHBoxLayout()
        self._statusbar_sys = QtWidgets.QLabel('')
        self._statusbar_usr = QtWidgets.QLabel('')
        hbox.addWidget(self._statusbar_sys)
        hbox.addWidget(self._statusbar_usr, 0, (QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter))

        # window resize grip
        self._grip = QtWidgets.QSizeGrip(self)
        hbox.addWidget(self._grip)

        self.layout.addLayout(hbox, len(self.parmList) + 3, 0)
        self.layout.setRowStretch(len(self.parmList) + 3, 0)

        # uid display
        # uid   = QtWidgets.QLabel("uid: "+str(self.node.getID()))
        # uid.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        # self.layout.addWidget(uid,len(self.parmList)+2,0)

        # instantiate the layout
        self.setLayout(self.layout)

        # run through all widget titles since each widget parent is now set.
        for parm in self.parmList:
            parm.setDispTitle()

        # instantiate the layout
        # self.setGeometry(50, 50, 300, 40)

        self.setTitle(node.getModuleName())

        self._starttime = 0
        self._startline = 0

    def initUI_return(self):
        return self._initUI_ret

    def getWidgets(self):
        # return a list of widgets
        return self.parmList

    def getWidgetNames(self):
        return list(self.parmDict.keys())

    def starttime(self):
        """Begin the timer for the node `Wall Time` calculation.

        Nodes store their own runtime, which is displayed in a tooltip when
        hovering over the node on the canvas (see :doc:`../ui`). Normally, each
        node reports the complete time it takes to run its :py:meth:`compute`
        function. However, a dev can use this method along with
        :py:meth:`endtime` to set the portion of :py:meth:`compute` to be used
        to calculate the `Wall Time`.
        """
        self._startline = inspect.currentframe().f_back.f_lineno
        self._starttime = time.time()

    def endtime(self, msg=''):
        """Begin the timer for the node `Wall Time` calculation.

        Nodes store their own runtime, which is displayed in a tooltip when
        hovering over the node on the canvas (see :doc:`../ui`). Normally, each
        node reports the complete time it takes to run its :py:meth:`compute`
        function. However, a dev can use this method along with
        :py:meth:`starttime` to set the portion of :py:meth:`compute` to be
        used to calculate the `Wall Time`. This method also prints the `Wall
        Time` to stdout, along with an optional additional message, using
        :py:meth:`gpi.logger.PrintLogger.node`).

        Args:
            msg (string): a message to be sent to stdout (using
                :py:meth:`gpi.logger.PrintLogger.node`) along with the `Wall
                Time`
        """
        ttime = time.time() - self._starttime
        eline = inspect.currentframe().f_back.f_lineno
        log.node(self.node.getName()+' - '+str(ttime)+'sec, between lines:'+str(self._startline)+'-'+str(eline)+'. '+msg)

    def stringifyExecType(self):
        if self.execType() is GPI_PROCESS:
            return " [Process]"
        elif self.execType() is GPI_THREAD:
            return " [Thread]"
        else:
            return " [App-Loop]"

    def setStatus_sys(self, msg):
        msg += self.stringifyExecType()
        self._statusbar_sys.setText(msg)

    def setStatus(self, msg):
        self._statusbar_usr.setText(msg)

    def execType(self):
        # default executable type
        # return GPI_THREAD
        return GPI_PROCESS  # this is the safest
        # return GPI_APPLOOP

    def setReQueue(self, val=False):  # NODEAPI
        # At the end of a nodeQueue, these tasked are checked for
        # more events.
        if self.node.inDisabledState():
            return
        if self.node.nodeCompute_thread.execType() == GPI_PROCESS:
            self.node.nodeCompute_thread.addToQueue(['setReQueue', val])
        else:
            self.node._requeue = val

    def reQueueIsSet(self):
        return self.node._requeue

    def getLabel(self):
        return self.label

    def moduleExists(self, name):
        """Give the user a simple module checker for the node validate
        function.
        """
        try:
            imp.find_module(name)
        except ImportError:
            return False
        else:
            return True

    def moduleValidated(self, name):
        """Provide a stock error message in the event a c/ext module cannot
        be found.
        """
        if not self.moduleExists(name):
            log.error("The \'" + name + "\' module cannot be found, compute() aborted.")
            return 1
        return 0

    def setLabelWidget(self, newlabel=''):
        self.wdglabel.setText(newlabel)

    def setLabel(self, newlabel=''):
        self.label = str(newlabel)
        self.updateTitle()
        self.node.updateOutportPosition()
        self.node.graph.scene().update(self.node.boundingRect())
        self.node.update()

    def openNodeDocumentation(self):
        self.doc_text_win.show()

        # setting the size only works if we have shown the widget
        # set the width based on some ideal width (max at 800px)
        docwidth = self.doc_text_win.document().idealWidth()
        lmargin = self.doc_text_win.contentsMargins().left()
        rmargin = self.doc_text_win.contentsMargins().right()
        scrollbar_width = 20 # estimate, scrollbar overlaps content otherwise
        total_width = min(lmargin + docwidth + rmargin + scrollbar_width, 800)
        self.doc_text_win.setFixedWidth(total_width)

        # set the height based on the content size
        docheight= self.doc_text_win.document().size().height()
        self.doc_text_win.setMinimumHeight(min(docheight, 200))
        self.doc_text_win.setMaximumHeight(docheight)

    def setDetailLabel(self, newDetailLabel='', elideMode='middle'):
        """Set an additional label for the node.

        This offers a way to programmatically set an additional label for a
        node (referred to as the `detail label`), which shows up underneath the
        normal node tile and label (as set wihtin the node menu). This is used
        by the `core` library to show file paths for the file reader/writer
        nodes, for example.

        Args:
            newDetailLabel (string): The detail label for the node (e.g. file
                path, operator, ...)
            elideMode ({'middle', 'left', 'right', 'none'}, optional): Method
                to use when truncating the detail label with an ellipsis.
        """
        self._detailLabel = str(newDetailLabel)
        self._detailElideMode = elideMode
        self.node.updateOutportPosition()

    def getDetailLabel(self):
        """Get the current node detail label.

        Returns:
            string: The node detail label, as set by :py:meth:`setDetailLabel`
        """
        return self._detailLabel

    def getDetailLabelElideMode(self):
        """How the detail label should be elided if it's too long:"""
        mode = self._detailElideMode
        qt_mode = QtCore.Qt.ElideMiddle
        if mode == 'left':
            qt_mode = QtCore.Qt.ElideLeft
        elif mode == 'right':
            qt_mode = QtCore.Qt.ElideRight
        elif mode == 'none':
            qt_mode = QtCore.Qt.ElideNone
        else: # default, mode == 'middle'
            qt_mode = QtCore.Qt.ElideMiddle

        return qt_mode

    # to be subclassed and reimplemented.
    def initUI(self):
        """Initialize the node UI (Node Menu).

        This method is intended to be reimplemented by the external node
        developer. This is where :py:class:`Port` and :py:class:`Widget`
        objects are added and initialized. This method is called whenever a
        node is added to the canvas.

        See :ref:`adding-widgets` and :ref:`adding-ports` for more detail.
        """

        # window
        #self.setWindowTitle(self.node.name)

        # IO Ports
        #self.addInPort('in1', str, obligation=REQUIRED)
        #self.addOutPort('out2', int)
        pass

    def windowActivationChange(self, test):
        self.node.graph.scene().makeOnlyTheseNodesSelected([self.node])

    def getSettings(self):  # NODEAPI
        """Wrap up all settings from each widget."""
        s = {}
        s['label'] = self.label
        s['parms'] = []
        for parm in self.parmList:
            s['parms'].append(copy.deepcopy(parm.getSettings()))
        return s

    def loadSettings(self, s):
        self.setLabelWidget(s['label'])

        # modify node widgets
        for parm in s['parms']:
            # the NodeAPI has instantiated the widget by name, this will
            # change the wdg-ID, however, this step is only dependend on
            # unique widget names.

            if not self.getWidget(parm['name']):
                log.warn("Trying to load settings; can't find widget with name: \'" + \
                    stw(parm['name']) + "\', skipping...")
                continue

            log.debug('Setting widget: \'' + stw(parm['name']) + '\'')
            try:
                self.modifyWidget_direct(parm['name'], **parm['kwargs'])
            except:
                log.error('Failed to set widget: \'' + stw(parm['name']) + '\'\n' + str(traceback.format_exc()))

            if parm['kwargs']['inport']:  # widget-inports
                self.addWidgetInPortByName(parm['name'])
            if parm['kwargs']['outport']:  # widget-outports
                self.addWidgetOutPortByName(parm['name'])


    def generateHelpText(self):
        """Gather the __doc__ string of the ExternalNode derived class,
        all the set_ methods for each attached widget, and any attached
        GPI-types from each of the ports.
        """
        if self._docText is not None:
            return self._docText

        # NODE DOC
        node_doc = "NODE: \'" + self.node._moduleName + "\'\n" + "    " + \
            str(self.__doc__)

        # WIDGETS DOC
        # parm_doc = ""  # contains parameter info
        wdg_doc = "\n\nAPPENDIX A: (WIDGETS)\n"  # generic widget ref info
        for parm in self.parmList:  # wdg order matters
            wdg_doc += "\n  \'" + parm.getTitle() + "\': " + \
                str(parm.__class__) + "\n"
            numSpaces = 8
            set_doc = "\n".join((numSpaces * " ") + i for i in str(
                parm.__doc__).splitlines())
            wdg_doc += set_doc + "\n"
            # set methods
            for member in dir(parm):
                if member.startswith('set_'):
                    wdg_doc += (8 * " ") + member + inspect.formatargspec(
                        *inspect.getargspec(getattr(parm, member)))
                    set_doc = str(inspect.getdoc(getattr(parm, member)))
                    numSpaces = 16
                    set_doc = "\n".join((
                        numSpaces * " ") + i for i in set_doc.splitlines())
                    wdg_doc += "\n" + set_doc + "\n"

        # PORTS DOC
        port_doc = "\n\nAPPENDIX B: (PORTS)\n"  # get the port type info
        for port in self.node.getPorts():
            typ = port.GPIType()
            port_doc += "\n  \'" + port.portTitle + "\': " + \
                str(typ.__class__) + "|" + str(type(port)) + "\n"
            numSpaces = 8
            set_doc = "\n".join((numSpaces * " ") + i for i in str(
                typ.__doc__).splitlines())
            port_doc += set_doc + "\n"

            # set methods
            for member in dir(typ):
                if member.startswith('set_'):
                    port_doc += (8 * " ") + member + inspect.formatargspec(
                        *inspect.getargspec(getattr(typ, member)))
                    set_doc = str(inspect.getdoc(getattr(typ, member)))
                    numSpaces = 16
                    set_doc = "\n".join((
                        numSpaces * " ") + i for i in set_doc.splitlines())
                    port_doc += "\n" + set_doc + "\n"

        # GETTERS/SETTERS
        getset_doc = "\n\nAPPENDIX C: (GETTERS/SETTERS)\n\n"
        getset_doc += (8 * " ") + "Node Initialization Setters: (initUI())\n\n"
        getset_doc += self.formatFuncDoc(self.addWidget)
        getset_doc += self.formatFuncDoc(self.addInPort)
        getset_doc += self.formatFuncDoc(self.addOutPort)
        getset_doc += (
            8 * " ") + "Node Compute Getters/Setters: (compute())\n\n"
        getset_doc += self.formatFuncDoc(self.getVal)
        getset_doc += self.formatFuncDoc(self.getAttr)
        getset_doc += self.formatFuncDoc(self.setAttr)
        getset_doc += self.formatFuncDoc(self.getData)
        getset_doc += self.formatFuncDoc(self.setData)

        self._docText = node_doc  # + wdg_doc + port_doc + getset_doc

        self.doc_text_win.setPlainText(self._docText)

        return self._docText

    def formatFuncDoc(self, func):
        """Generate auto-doc for passed func obj."""
        numSpaces = 24
        fdoc = inspect.getdoc(func)
        set_doc = "\n".join((
            numSpaces * " ") + i for i in str(fdoc).splitlines())
        rdoc = (16 * " ") + func.__name__ + \
            inspect.formatargspec(*inspect.getargspec(func)) \
            + "\n" + set_doc + "\n\n"
        return rdoc

    def printWidgetValues(self):
        # for debugging
        for parm in self.parmList:
            log.debug("widget: " + str(parm))
            log.debug("value: " + str(parm.get_val()))

    # abstracting IF for user
    def addInPort(self, title=None, type=None, obligation=REQUIRED,
                  menuWidget=None, cyclic=False, **kwargs):
        """Add an input port to the node.

        Input ports collect data from other nodes for use/processing within a
        node compute function. Ports may only be added in the :py:meth:`initUI`
        routine. Data at an input node can be accessed using
        :py:meth:`getData`, but is typically read-only.

        Args:
            title (str): port title (shown in tooltips) and unique identifier
            type (str): class name of extended type (e.g. ``np.complex64``)
            obligation: ``gpi.REQUIRED`` (default) or ``gpi.OPTIONAL``
            menuWidget: `for internal use`, devs should leave as default
                ``None``
            cyclic (bool): whether the port should allow reverse-flow from
                downstream nodes (default is ``False``)
            kwargs: any set_<arg> method belonging to the ``GPIDefaultType``
                derived class
        """
        self.node.addInPort(title, type, obligation, menuWidget, cyclic, **kwargs)
        self.node.update()

    # abstracting IF for user
    def addOutPort(self, title=None, type=None, obligation=REQUIRED,
                   menuWidget=None, **kwargs):
        """Add an output port to the node.

        Output nodes provide a conduit for passing data to downstream nodes.
        Ports may only be added in the :py:meth:`initUI` routine.  Data at an
        output port can be accessed using :py:meth:`setData` and
        :py:meth:`getData` from within :py:meth:`validate` and
        :py:meth:`compute`.

        Args:
            title (str): port title (shown in tooltips) and unique identifier
            type (str): class name of extended type (e.g. ``np.float32``)
            obligation: ``gpi.REQUIRED`` (default) or ``gpi.OPTIONAL``
            menuWidget: `for internal use`, debs should leave as default
                ``None``
            kwargs: any set_<arg> method belonging to the ``GPIDefaultType``
                derived class
        """
        self.node.addOutPort(title, type, obligation, menuWidget, **kwargs)
        self.node.update()

    def addWidgetInPortByName(self, title):
        wdg = self.findWidgetByName(title)
        self.addWidgetInPort(wdg)

    def addWidgetInPortByID(self, wdgID):
        wdg = self.findWidgetByID(wdgID)
        self.addWidgetInPort(wdg)

    def addWidgetInPort(self, wdg):
        self.addInPort(title=wdg.getTitle(), type=wdg.getDataType(),
                       obligation=OPTIONAL, menuWidget=wdg)

    def removeWidgetInPort(self, wdg):
        port = self.findWidgetInPortByName(wdg.getTitle())
        self.node.removePortByRef(port)

    def addWidgetOutPortByID(self, wdgID):
        wdg = self.findWidgetByID(wdgID)
        self.addWidgetOutPort(wdg)

    def addWidgetOutPortByName(self, title):
        wdg = self.findWidgetByName(title)
        self.addWidgetOutPort(wdg)

    def addWidgetOutPort(self, wdg):
        self.addOutPort(
            title=wdg.getTitle(), type=wdg.getDataType(), menuWidget=wdg)

    def removeWidgetOutPort(self, wdg):
        port = self.findWidgetOutPortByName(wdg.getTitle())
        self.node.removePortByRef(port)

    def addWidget(self, wdg=None, title=None, **kwargs):
        """Add a widget to the node UI.

        Args:
            wdg (str): The widget class name (see :doc:`widgets` for a list of
                built-in widget classes)
            title (str): A unique name for the widget, and title shown in the
                Node Menu
            kwargs: Additional arguments passed to the set_<arg> methods
                specific to the chosen widget class
        """

        if (wdg is None) or (title is None):
            log.critical("addWidget(): widgets need a title" \
                + " AND a wdg-str! Aborting.")
            return

        # check existence first
        if self.node.titleExists(title):
            log.critical("addWidget(): Widget title \'" + str(title) \
                + "\' is already in use! Aborting.")
            return

        ypos = len(self.parmList)

        # try widget def's packaged with node def first
        wdgGroup = None

        # first see if the node code contains the widget def
        wdgGroup = self.node.item.getWidget(wdg)

        # get widget from standard gpi widgets
        if wdgGroup is None:
            if hasattr(BUILTIN_WIDGETS, wdg):
                wdgGroup = getattr(BUILTIN_WIDGETS, wdg)

        # if its still not found then its a big deal
        if wdgGroup is None:
            log.critical("\'" + wdg + "\' widget not found.")
            raise Exception("Widget not found.")

        # instantiate if not None
        else:
            wdgGroup = wdgGroup(title)

        wdgGroup.setNodeName(self.node.getModuleName())
        wdgGroup._setNodeLabel(self.label)

        wdgGroup.valueChanged.connect(lambda: self.wdgEvent(title))
        wdgGroup.portStateChange.connect(lambda: self.changePortStatus(title))
        wdgGroup.returnWidgetToOrigin.connect(self.returnWidgetToNodeMenu)
        self.wdglabel.textChanged.connect(wdgGroup._setNodeLabel)

        # add to menu layout
        self.layout.addWidget(wdgGroup, ypos, 0)
        self.parmList.append(wdgGroup)
        self.parmDict[title] = wdgGroup  # the new way

        self.modifyWidget_direct(title, **kwargs)

    def returnWidgetToNodeMenu(self, wdgid):
        # find widget by id
        wdgid = int(wdgid)
        ind = 0
        for parm in self.parmList:
            if wdgid == parm.get_id():
                self.layout.addWidget(parm, ind, 0)
                if self.node.isMarkedForDeletion():
                    parm.hide()
                else:
                    parm.show()
                break
            ind += 1

    def blockWdgSignals(self, val):
        """Block all signals, especially valueChanged."""
        for parm in self.parmList:
            parm.blockSignals(val)

    def blockSignals_byWdg(self, title, val):
        # get the over-reporting widget by name and block it
        self.parmDict[title].blockSignals(val)

    def changePortStatus(self, title):
        """Add a new in- or out-port tied to this widget."""
        log.debug("changePortStatus: " + str(title))
        wdg = self.findWidgetByName(title)

        # make sure the port is either added or will be added.
        if wdg.inPort_ON:
            if self.findWidgetInPortByName(title) is None:
                self.addWidgetInPort(wdg)
        else:
            if self.findWidgetInPortByName(title):
                self.removeWidgetInPort(wdg)
        if wdg.outPort_ON:
            if self.findWidgetOutPortByName(title) is None:
                self.addWidgetOutPort(wdg)
        else:
            if self.findWidgetOutPortByName(title):
                self.removeWidgetOutPort(wdg)

    def findWidgetByName(self, title):
        for parm in self.parmList:
            if parm.getTitle() == title:
                return parm

    def findWidgetByID(self, wdgID):
        for parm in self.parmList:
            if parm.id() == wdgID:
                return parm

    def findWidgetInPortByName(self, title):
        for port in self.node.inportList:
            if port.isWidgetPort():
                if port.portTitle == title:
                    return port

    def findWidgetOutPortByName(self, title):
        for port in self.node.outportList:
            if port.isWidgetPort():
                if port.portTitle == title:
                    return port

    def modifyWidget_setter(self, src, kw, val):
        sfunc = "set_" + kw
        if hasattr(src, sfunc):
            func = getattr(src, sfunc)
            func(val)
        else:
            try:
                ttl = src.getTitle()
            except:
                ttl = str(src)

            log.critical("modifyWidget_setter(): Widget \'" + stw(ttl) \
                + "\' doesn't have attr \'" + stw(sfunc) + "\'.")

    def modifyWidget_direct(self, pnumORtitle, **kwargs):
        src = self.getWidget(pnumORtitle)

        for k, v in list(kwargs.items()):
            if k != 'val':
                self.modifyWidget_setter(src, k, v)

        # set 'val' last so that bounds don't cause a temporary conflict.
        if 'val' in kwargs:
            self.modifyWidget_setter(src, 'val', kwargs['val'])

    def modifyWidget_buffer(self, title, **kwargs):
        """GPI_PROCESSes have to use buffered widget attributes to effect the
        same changes to attributes during compute() as with GPI_THREAD or
        GPI_APPLOOP.
        """
        src = self.getWdgFromBuffer(title)

        try:
            for k, v in list(kwargs.items()):
                src['kwargs'][k] = v
        except:
            log.critical("modifyWidget_buffer() FAILED to modify buffered attribute")


    def lockWidgetUpdates(self):
        self._widgetUpdateMutex_locked = True

    def unlockWidgetUpdates(self):
        self._widgetUpdateMutex_locked = False

    def widgetsUpdatesLocked(self):
        return self._widgetUpdateMutex_locked

    def wdgEvent(self, title):
        # Captures all valueChanged events from widgets.
        # 'title' is for interrogating who changed in compute().

        # Once the event status has been set, disable valueChanged signals
        # until the node has successfully completed.
        # -this was b/c sliders etc. were causing too many signals resulting
        # in a recursion overload to this function.
        #self.blockWdgSignals(True)
        self.blockSignals_byWdg(title, True)

        if self.node.hasEventPending():
            self.node.appendEvent({GPI_WIDGET_EVENT: title})
            return
        else:
            self.node.setEventStatus({GPI_WIDGET_EVENT: title})

        # Can start event from any applicable 'check' transition
        if self.node.graph.inIdleState():
            self.node.graph._switchSig.emit('check')

    # for re-drawing the menu title to match module/node instance
    def updateTitle(self):
        # Why is this trying the scrollArea? isn't it always a scroll???
        if self.label == '':
            try:
                self.node._nodeIF_scrollArea.setWindowTitle(self.node.name)
            except:
                self.setWindowTitle(self.node.name)
        else:
            try:
                augtitle = self.node.name + ": " + self.label
                self.node._nodeIF_scrollArea.setWindowTitle(augtitle)
            except:
                augtitle = self.node.name + ": " + self.label
                self.setWindowTitle(augtitle)

    def setTitle(self, title):
        self.node.name = title
        self.updateTitle()

    # Queue actions for widgets and ports
    def setAttr(self, title, **kwargs):
        """Set specific attributes of a given widget.

        This method may be used to set attributes of any widget during any of
        the core node functions: :py:meth:`initUI`, :py:meth:`validate`, or
        :py:meth:`compute`.

        Args:
            title (str): the widget name (unique identifier)
            kwargs: args corresponding to the ``get_<arg>`` methods of the
                widget class. See :doc:`widgets` for the list of built-in
                widgets and associated attributes.
        """
        try:
            # start = time.time()
            # either emit a signal or save a queue
            if self.node.inDisabledState():
                return
            if self.node.nodeCompute_thread.execType() == GPI_PROCESS:
                # PROCESS
                self.node.nodeCompute_thread.addToQueue(['modifyWdg', title, kwargs])
                #self.modifyWidget_buffer(title, **kwargs)
            elif self.node.nodeCompute_thread.execType() == GPI_THREAD:
                # THREAD
                # can't modify QObjects in thread, but we can send a signal to do
                # so
                self.modifyWdg.emit(title, kwargs)
            else:
                # APPLOOP
                self.modifyWidget_direct(str(title), **kwargs)

        except:
            raise GPIError_nodeAPI_setAttr('self.setAttr(\''+stw(title)+'\',...) failed in the node definition, check the widget name, attribute name and attribute type().')

        # log.debug("modifyWdg(): time: "+str(time.time() - start)+" sec")

    def allocArray(self, shape=(1,), dtype=np.float32, name='local'):
        """return a shared memory array if the node is run as a process.
            -the array name needs to be unique
        """
        if self.node.nodeCompute_thread.execType() == GPI_PROCESS:
            buf, shd = DataProxy()._genNDArrayMemmap(shape, dtype, self.node.getID(), name)

            if shd is not None:
                # saving the reference id allows the node developer to decide
                # on the fly if the preallocated array will be used in the final
                # setData() call.
                self.shdmDict[str(id(buf))] = shd.filename

            return buf
        else:
            return np.ndarray(shape, dtype=dtype)

    def setData(self, title, data):
        """Set the data at an :py:class:`OutPort`.

        This is typically called in :py:meth:`compute` to set data at an output
        port, making the data available to downstream nodes.
        :py:class:`InPort` ports are read-only, so this method should only be used
        with :py:class:`OutPort` ports.

        Args:
            title (str): name of the port to send the object reference
            data: any object corresponding to a ``GPIType`` class allowed by
                this port
        """
        try:
            # start = time.time()
            # either set directly or save a queue
            if self.node.inDisabledState():
                return
            if self.node.nodeCompute_thread.execType() == GPI_PROCESS:

                #  numpy arrays
                if type(data) is np.memmap or type(data) is np.ndarray:
                    if str(id(data)) in self.shdmDict: # pre-alloc
                        s = DataProxy().NDArray(data, shdf=self.shdmDict[str(id(data))], nodeID=self.node.getID(), portname=title)
                    else:
                        s = DataProxy().NDArray(data, nodeID=self.node.getID(), portname=title)

                    # for split objects to pass thru individually
                    # this will be a list of DataProxy objects
                    if type(s) is list:
                        for i in s:
                            self.node.nodeCompute_thread.addToQueue(['setData', title, i])
                    # a single DataProxy object
                    else:
                        self.node.nodeCompute_thread.addToQueue(['setData', title, s])

                # all other non-numpy data that are pickleable
                else:
                    # PROCESS output other than numpy
                    self.node.nodeCompute_thread.addToQueue(['setData', title, data])
            else:
                # THREAD or APPLOOP
                self.node.setData(title, data)
                # log.debug("setData(): time: "+str(time.time() - start)+" sec")

        except:
            print((str(traceback.format_exc())))
            raise GPIError_nodeAPI_setData('self.setData(\''+stw(title)+'\',...) failed in the node definition, check the output name and data type().')

    def getData(self, title):
        """Get the data from a :py:class:`Port` for this node.

        Usually this is used to get input data from a :py:class:`InPort`,
        though in some circumstances (e.g. in the `Glue` node) it may be used
        to get data from an :py:class:`OutPort`. This method is available for
        devs to use in :py:meth:`validate` and :py:meth:`compute`.

        Args:
            title (str): the name of the GPI :py:class:`Port` object
        Returns:
            Data from the :py:class:`Port` object. The return will have a type
            corresponding to a ``GPIType`` class allowed by this port.
        """
        try:
            port = self.node.getPortByNumOrTitle(title)
            if isinstance(port, InPort):
                data = port.getUpstreamData()
                if type(data) is np.ndarray:
                    # don't allow users to change original array attributes
                    # that aren't protected by the 'writeable' flag
                    buf = np.frombuffer(data.data, dtype=data.dtype)
                    buf.shape = tuple(data.shape)
                    return buf
                else:
                    return data

            elif isinstance(port, OutPort):
                return port.data
            else:
                raise Exception("getData", "Invalid Port Title")
        except:
            raise GPIError_nodeAPI_getData('self.getData(\''+stw(title)+'\') failed in the node definition check the port name.')

    def getInPort(self, pnumORtitle):
        return self.node.getInPort(pnumORtitle)

    def getOutPort(self, pnumORtitle):
        return self.node.getOutPort(pnumORtitle)

############### DEPRECATED NODE API
# TTD v0.3

    def getAttr_fromWdg(self, title, attr):
        """title = (str) wdg-class name
        attr = (str) corresponds to the get_<arg> of the desired attribute.
        """
        log.warn('The \'getAttr_fromWdg()\' function is deprecated, use \'getAttr()\' instead.  '+str(self.node.getFullPath()))
        return self.getAttr(title, attr)

    def getVal_fromParm(self, title):
        """Returns get_val() from wdg-class (see getAttr()).
        """
        log.warn('The \'getVal_fromParm()\' function is deprecated, use \'getVal()\' instead.  '+str(self.node.getFullPath()))
        return self.getVal(title)

    def getData_fromPort(self, title):
        """title = (str) the name of the InPort.
        """
        log.warn('The \'getData_fromPort()\' function is deprecated, use \'getData()\' instead.  '+str(self.node.getFullPath()))
        return self.getData(title)

    def setData_ofPort(self, title, data):
        """title = (str) name of the OutPort to send the object reference.
        data = (object) any object corresponding to a GPIType class.
        """
        log.warn('The \'setData_ofPort()\' function is deprecated, use \'setData()\' instead.  '+str(self.node.getFullPath()))
        self.setData(title, data)

    def modifyWidget(self, title, **kwargs):
        """title = (str) the corresponding widget name.
        kwargs = args corresponding to the get_<arg> methods of the wdg-class.
        """
        log.warn('The \'modifyWidget()\' function is deprecated, use \'setAttr()\' instead.  '+str(self.node.getFullPath()))
        self.setAttr(title, **kwargs)

    def getEvent(self):
        """Allow node developer to get information about what event has caused
        the node to run."""
        log.warn('The \'getEvent()\' function is deprecated, use \'getEvents()\' (its the plural form). '+str(self.node.getFullPath()))
        return self.node.getPendingEvent()

    def portEvent(self):
        """Specifically check for a port event."""
        log.warn('The \'portEvent()\' function is deprecated, use \'portEvents()\' (its the plural form). '+str(self.node.getFullPath()))
        if GPI_PORT_EVENT in self.getEvent():
            return self.getEvent()[GPI_PORT_EVENT]
        return None

    def widgetEvent(self):
        """Specifically check for a wdg event."""
        log.warn('The \'widgetEvent()\' function is deprecated, use \'widgetEvents()\' (its the plural form). '+str(self.node.getFullPath()))
        if GPI_WIDGET_EVENT in self.getEvent():
            return self.getEvent()[GPI_WIDGET_EVENT]
        return None
############### DEPRECATED NODE API

    def getEvents(self):
        """Get information about events that caused the node to run.

        Returns a dictionary containing names of widgets and ports that have
        changed values since the last time the node ran. Additionally contains
        information regarding ``init`` or ``requeue`` events trigered by GPI's
        node-evaluation routines.

        `Note: events from widget-ports count as both widget and port events.`

        Returns:
            dict: all events accumulated since the node was last run

                The event dictionary contains four key:value pairs:
                    * ``GPI_WIDGET_EVENT`` : `set(widget_titles (string)`)
                    * ``GPI_PORT_EVENT`` : `set(port_titles (string)`)
                    * ``GPI_INIT_EVENT`` : ``True`` or ``False``
                    * ``GPI_REQUEUE_EVENT`` : ``True`` or ``False``
        """
        return self.node.getPendingEvents().events

    def portEvents(self):
        """Specifically check for port events.

        Get the names (unique identifier strings) of any ports that have
        changed data since the last run.

        `Note: events from widget-ports count as both widget and port events.`

        Returns:
            set: names (strings) of all ports modified since the node last ran
        """
        return self.node.getPendingEvents().port

    def widgetEvents(self):
        """Specifically check for a widget events.

        Get the names (unique identifier strings) of any widgets that have
        changed data since the last run.

        `Note: events from widget-ports count as both widget and port events.`

        Returns:
            set: names (strings) of all widgets modified since the node last
                ran
        """
        return self.node.getPendingEvents().widget

    def widgetMovingEvent(self, wdgid):
        """Called when a widget drag is being initiated.
        """
        pass

    def getWidgetByID(self, wdgID):
        for parm in self.parmList:
            if parm.id() == wdgID:
                return parm
        log.critical("getWidgetByID(): Cannot find widget id:" + str(wdgID))

    def getWidget(self, pnum):
        """Returns the widget desc handle and position number"""
        # fetch by widget number
        if type(pnum) is int:
            if (pnum < 0) or (pnum >= len(self.parmList)):
                log.error("getWidget(): Target widget out of range: " + str(pnum))
                return
            src = self.parmList[pnum]
        # fetch by widget title
        elif type(pnum) is str:
            src = None
            cnt = 0
            for parm in self.parmList:
                if parm.getTitle() == pnum:
                    src = parm
                    pnum = cnt  # change pnum back to int
                cnt += 1
            if src is None:
                log.error("getWidget(): Failed to find widget: \'" + stw(pnum) + "\'")
                return
        else:
            log.error("getWidget(): Widget identifier must be" + " either int or str")
            return
        return src

    def bufferParmSettings(self):
        """Get list of parms (dict) in self.parmSettings['parms'].
        Called by GPI_PROCESS functor to capture all widget settings needed
        in compute().
        """
        self.parmSettings = self.getSettings()

    def getWdgFromBuffer(self, title):
        # find a specific widget by title
        for wdg in self.parmSettings['parms']:
            if 'name' in wdg:
                if wdg['name'] == title:
                    return wdg


    def getVal(self, title):
        """Returns the widget value.

        Each widget class has a corresponding "main" value. This method will
        return the value of the widget, by passing along the return from its
        `get_val()` method.

        Returns:
            The widget value. The type of the widget value is defined by the
            widget class. See :doc:`widgets` for value types for the built-in
            widget classes.
        """
        try:
            # Fetch widget value by title
            if self.node.inDisabledState():
                return

            if self.node.nodeCompute_thread.execType() == GPI_PROCESS:
                return self.getAttr(title, 'val')

            # threads and main loop can access directly
            return self.getAttr(title, 'val')

        except:
            print(str(traceback.format_exc()))
            raise GPIError_nodeAPI_getVal('self.getVal(\''+stw(title)+'\') failed in the node definition, check the widget name.')

    def getAttr(self, title, attr):
        """Get a specific attribute value from a widget.

        This returns the value of a specific attribute of a widget. Widget
        attributes may be modified by the user manipulating the Node Menu,
        during widget creation using the `kwargs` in :py:meth:`addWidget`, or
        programmatically by :py:meth:`setAttr`.

        Args:
            title (str): The widget name (unique identifier)
            attr (str): The desired attribute
        Returns:
            The desired widget attribute. This value is retrieved by calling
            ``get_<attr>`` on the indicated widget. See :doc:`widgets` for a
            list of attributes for the buil-in widget classes.
        """
        try:
            # Fetch widget value by title
            if self.node.inDisabledState():
                return

            if self.node.nodeCompute_thread.execType() == GPI_PROCESS:
                wdg = self.getWdgFromBuffer(title)
                return self._getAttr_fromWdg(wdg, attr)

            # threads and main loop can access directly
            wdg = self.getWidget(title)
            return self._getAttr_fromWdg(wdg, attr)

        except:
            print(str(traceback.format_exc()))
            raise GPIError_nodeAPI_getAttr('self.getAttr(\''+stw(title)+'\',...) failed in the node definition, check widget name and attribute name.')

    def _getAttr_fromWdg(self, wdg, attr):
        try:
            # input is either a dict or a wdg instance.
            if isinstance(wdg, dict):  # buffered values
                if attr in wdg['kwargs']:
                    return wdg['kwargs'][attr]
                else:
                    # build error msg
                    title = wdg['name']
                    funame = attr
            else:  # direct widget access
                funame = 'get_' + attr
                if hasattr(wdg, funame):
                    return getattr(wdg, funame)()
                else:
                    try:
                        title = wdg.getTitle()
                    except:
                        title = "\'no name\'"
            log.critical("_getAttr(): widget \'" + stw(title) + "\' has no attr \'" + stw(funame) + "\'.")
            #return None
            raise GPIError_nodeAPI_getAttr('_getAttr() failed for widget \''+stw(title)+'\'')
        except:
            #log.critical("_getAttr_fromWdg(): Likely the wrong input arg type.")
            #raise
            print(str(traceback.format_exc()))
            raise GPIError_nodeAPI_getAttr('_getAttr() failed for widget \''+stw(title)+'\'')

    def validate(self):
        """The pre-compute validation step.

        This function is intended to be reimplemented by the external node
        developer. Here the developer can access widget and port data (see
        :ref:`accessing-widgets` and :ref:`accessing-ports`) to perform
        validation checks before :py:meth:`compute` is called.

        Returns:
            An integer corresponding to the result of the validation:
                0: The node successfully passed validation

                1: The node failed validation, compute will not be called and
                the canvas will be paused
        """
        log.debug("Default module validate().")
        return 0

    def compute(self):
        """The module compute routine.

        This function is intended to be reimplemented by the external node
        developer. This is where the main computation of the node is performed.
        The developer has full access to the widget and port data (see
        :ref:`accessing-widgets` and :ref:`accessing-ports`).

        Returns:
            An integer corresponding to the result of the computation:
                0: Compute completed successfully

                1: Compute failed in some way, the canvas will be paused
        """
        log.debug("Default module compute().")
        return 0

    def post_compute_widget_update(self):
        # reset any widget that requires it (i.e. PushButton)
        for parm in self.parmList:
            if hasattr(parm, 'set_reset'):
                parm.set_reset()


#!/usr/bin/env python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: Encapsulate numba decorators so the code can be used w/ or w/o numba.
 

from .logger import manager
# start logger for this module
log = manager.getLogger(__name__)

_USE_NUMBA = True

try:
    if not _USE_NUMBA:
        raise ImportError()  
    import numba
    from numba import autojit, jit  #, prange

    log.info('Numba version '+str(numba.__version__) +' found.')

except:
    log.info('Numba failed to load, using non-jit enhanced functions.')
    
    # use the original func identity of numba fails to load
    def autojit(func):
        return func

    def jit(arg):
        def dec(func):
            return func
        return dec



#!/opt/gpi/bin/python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


"""A simple test script for the PyFI template module.
"""

import os
import sys
import time
import numpy as np
import numpy.linalg as npl

import template as temp

import math
from math import sqrt




# PYCALLABLE
A = np.array(list(range(4)), dtype=np.float32)
A.shape = [2,2]
print("A:", A)
print("Ainv:", npl.pinv(A))

x = np.array(list(range(24)), dtype=np.float32)
x.shape = [4,3,2]

print(x)

print("matmult (PYTHON)")
print("A: ", A)
print("AdotA: ", np.dot(A,A))

st = time.time()
aarr, c = temp.IFtest2(x)
print("IFtest time: ", time.time()-st)

print(c)

print("aarr:", aarr)


temp.math_double()
temp.math_float()
temp.math_int()




# testing the PyFI interface.
def test(thetest, name):
    if thetest:
        print('\t-'+name+' passed')
        return 1
    else:
        print('\t-'+name+' failed')
        return 0




print("\n\nInterface testing...")
mi = 111
mf = 111.111
ms = 'hello from python'

acf = np.array([1 + 2j, 2, 3], dtype=np.complex64)
ad  = np.array([111.111, 22.22, 3.3], dtype=np.float64)
al  = np.array([7], dtype=np.int64)

# no keywords
# check KW defaults
preout, insize, oi, of, okf, oks, os, nchk, cfa, la, kcfa, kda, kla = temp.IFtest(mi, mf, ms, acf, ad, al)
cnt = 0
tot = 0

# tests
cnt += test((preout == 8.1).all(), 'preout')
tot += 1
cnt += test(oi == (1+111), 'myoutint')
tot += 1
cnt += test(of == mf, 'myfloat')
tot += 1
cnt += test(okf == -0.6, 'mykwfloat')
tot += 1
cnt += test(oks == '<<< your ad here >>>', 'mykwstring')
tot += 1
cnt += test(os == 'c++ generated string', 'myoutstring') 
tot += 1
cnt += test(nchk == 1, 'nullCheckPassed') 
tot += 1
cnt += test((insize == np.multiply(ad,ad)).all(), 'nullCheckPassed') 
tot += 1
cnt += test((cfa == acf).all(), 'mycfarr') 
tot += 1
cnt += test((la == al).all(), 'mylarr') 
tot += 1
cnt += test((kcfa == 1.0+1j).all(), 'mykwcfarr default') 
tot += 1
cnt += test((kda == 0.0).all(), 'mykwdarr default') 
tot += 1
cnt += test((kla == 7).all(), 'mykwlarr default') 
tot += 1

# test keywords
preout, insize, oi, of, okf, oks, os, nchk, cfa, la, kcfa, kda, kla = temp.IFtest(mi, mf, ms, acf, ad, al, mykwcfarr=acf, mykwdarr=ad, mykwlarr=al)

# tests
cnt += test((kcfa == acf).all(), 'mykwcfarr') 
tot += 1
cnt += test((kda == ad).all(), 'mykwdarr') 
tot += 1
cnt += test((kla == al).all(), 'mykwlarr') 
tot += 1

print('\nTest Summary:')
if cnt == tot:
    print('\tSUCCESS: '+str(cnt)+'/'+str(tot))
else:
    print('\tFAILED: '+str(cnt)+'/'+str(tot))



import pathlib
from setuptools import setup
from setuptools import find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='gpilab',
    version='1.4.4',
    description="Graphical Programming Interface",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/gpilab/framework",
    author="AbdulRahman Alfayad",
    author_email="alfayad.abdulrahman@mayo.edu",
    license="GNU",
    packages=find_packages(),
    install_requires=[
        "cycler==0.11.0",
        "dill==0.3.4",
        "fonttools==4.28.5",
        "grpcio==1.43.0",
        "grpcio-tools==1.43.0",
        "h5py==3.6.0",
        "kiwisolver==1.3.2",
        "matplotlib==3.5.1",
        "multiprocess==0.70.12.2",
        "numpy==1.21.5",
        "packaging==21.3",
        "pathos==0.2.8",
        "Pillow==9.0.1",
        "pox==0.3.0",
        "ppft==1.6.6.4",
        "protobuf==3.19.1",
        "psutil==5.8.0",
        "PyOpenGL==3.1.5",
        "pyparsing==3.0.6",
        "PyQt5==5.15.6",
        "PyQt5-Qt5==5.15.2",
        "PyQt5-sip==12.9.0",
        "pyqtgraph==0.12.3",
        "python-dateutil==2.8.2",
        "qimage2ndarray==1.8.3",
        "QtPy==2.0.0",
        "scipy==1.7.3",
        "six==1.16.0",
        "gpi_core"
    ],
    include_package_data=True,
    python_requires='>=3.7',
    scripts=['bin/gpi', 'bin/gpi_make', 'bin/gpi_update', 'bin/gpi_init', 'bin/gpi.cmd', 'bin/gpi_make.cmd']
)


#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

'''This module is an extension for handling specific data types such as
Numpy-arrays. '''

import os
import hashlib
import numpy as np

# gpi
from .defines import GPI_SHDM_PATH
from .logger import manager
from .sysspecs import Specs

# start logger for this module
log = manager.getLogger(__name__)


# List all types that are handled. This tells the deserializing side what to do
class ProxyType(object):
    null = -1
    np_ndarray = 0
    np_memmap = 1
    segmented = 2

class DataProxy(dict):
    '''Holds all file descriptor information for any object that is
    serializeable.  The method functions facilitate serialization and
    deserialization on either side of the Proxy-Manager.

    NUMPY:
        Numpy arrays must be segments that are smaller than 2GiB (2^31 bytes).

    NUMPY-MMAP:
        MMAP file descriptors are passed through the proxy only if there are
        enough available resources (i.e. rlimit).
    '''
    def __init__(self):
        super(DataProxy, self).__init__()
        #self['proxy_type'] = ProxyType.null

    def getSHMF(self, nodeID, name='local'):
        '''return a unique shared mem handle for this gpi instance, node and port.
        '''
        # make sure the user supplied string is a unique, consistent and valid filename
        hsh = hashlib.md5(str(name).encode('utf8')).hexdigest()

        # add a little salt with the random int generator - this will just grow
        # the ports don't keep track of these file names for cleanup
        #hsh = hashlib.md5(str(name)+str(np.random.randint(0,999))).hexdigest()

        return os.path.join(GPI_SHDM_PATH, str(hsh)+'_'+str(nodeID))

    def isSegmented(self):
        return self['proxy_type'] == ProxyType.segmented

    # select the correct proxy data for np-ndarrays and memmaps
    def NDArray(self, data, shdf=None, nodeID=None, portname=None):

        # if the user creates a memmapped numpy w/o using allocArray()
        if type(data) is np.memmap and data.filename is not None:
            # it's a *real* np.memmap
            self._setNDArrayMemmapFromNDArrayMemmap(data)

        # if the user is using an ndarray interface directly
        else:

            # if the user creates a memmapped numpy using allocArray()
            if shdf is not None:
                self._setNDArrayMemmapFromWrappedNDarrayMemmap(data, shdf)

            # normal numpy arrays
            else:

                # if the array is small then just send it directly instead of
                # using up a file handle
                if data.nbytes < 2**25: # 32MiB 
                    self._setNDArrayFromNDArray(data)

                # we're too close to the open file limit so start using segmented proxy
                elif Specs.openFileLimitThresh():
                    return self._genNDArraySegmentsFromNDArray(data)
            
                # in the normal case we'll use memmap to pass data.
                else:
                    self._setNDArrayMemmapFromNDArray(data, nodeID, portname)
        return self

    # no tricks just pass the np ndarray directly
    def _setNDArrayFromNDArray(self, data):
        self['proxy_type'] = ProxyType.np_ndarray
        self['data'] = data

    # np ndarray segment
    def _setNDArraySegmentFromNDArray(self, seg, oshape, no, total, did):
        self['proxy_type'] = ProxyType.segmented
        self['seg_type'] = ProxyType.np_ndarray
        self['id'] = did
        self['seg'] = seg 
        self['oshape'] = oshape
        self['no.'] = no
        self['total'] = total
        return self

    # if the process is out of file handles or the byte size of the array is
    # below the threshold, then use the segmented approach
    # returns a list of DataProxy objects.
    def _genNDArraySegmentsFromNDArray(self, data):

        log.info("------ SPLITTING LARGE NPY ARRAY >1GiB")
        div = int(data.nbytes/(2**30)) + 1 # 1GiB 

        oshape = list(data.shape)
        fshape = [np.prod(data.shape)]
        if not data.flags['C_CONTIGUOUS']:
            log.warn('Output array is not contiguous, forcing contiguity.')
            data = np.ascontiguousarray(data)
        data.shape = fshape  # flatten
        segs = np.array_split(data, div)
        did = id(data)

        buf = []
        cnt = 0
        tot = len(segs)
        for seg in segs:
            buf.append(DataProxy()._setNDArraySegmentFromNDArray(seg, oshape, cnt, tot, did))
            cnt += 1
        return buf

    # assemble all the numpy chunks into one array and return the array
    def _assembleNDArraySegments(self, segments):
        log.info("_assembleNDArraySegments(): ------ APPENDING LARGE NPY ARRAY SEGMENTS")

        if len(segments) != segments[0]['total']:
            log.error('Failed to proxy all numpy array segments. Aborting.')
            return

        # order the segments based on their 'no.'
        segments = sorted(segments, key=lambda d: d['no.'])

        # gather array segments and reshape NPY array
        segs = [s['seg'] for s in segments]
        lrgNPY = np.concatenate(segs)
        lrgNPY.shape = segments[0]['oshape']
        return lrgNPY

    # if an np-ndarray is passed then copy it to an np-memmap
    def _setNDArrayMemmapFromNDArray(self, data, nodeID, portname):
        self['proxy_type'] = ProxyType.np_memmap
        self['shape'] = tuple(data.shape)
        self['dtype'] = data.dtype
        self['shdf'] = self.getSHMF(nodeID, portname)
        fp = np.memmap(self['shdf'], dtype=data.dtype, mode='w+', shape=self['shape'])
        fp[:] = data[:] # full copy

    # if the np-memmap is already generated and passed directly then just copy
    # the relevant information
    def _setNDArrayMemmapFromNDArrayMemmap(self, data):
        self['proxy_type'] = ProxyType.np_memmap
        self['shape'] = tuple(data.shape)
        self['shdf'] = data.filename
        self['dtype'] = data.dtype

    # if a numpy array is wrapping a memmap'd array then pass the name
    def _setNDArrayMemmapFromWrappedNDarrayMemmap(self, data, shdf):
        self['proxy_type'] = ProxyType.np_memmap
        self['shape'] = tuple(data.shape)
        self['dtype'] = data.dtype
        self['shdf'] = shdf

    # create and return an np-ndarray wrapped memmap
    # return handles to both the wrapped and memmap'd data
    def _genNDArrayMemmap(self, shape=(1,), dtype=np.float32, nodeID=0, portname='local'):

        # too close to the open file limit so just give the user a normal array
        if Specs.openFileLimitThresh():
            log.warn("Maxed out file handles, pre-alloc will be ndarray...")
            return np.ndarray(shape, dtype=dtype), None

        fn = self.getSHMF(nodeID, portname)
        shd = np.memmap(fn, dtype=dtype, mode='w+', shape=tuple(shape))
        buf = np.frombuffer(shd.data, dtype=shd.dtype)
        buf.shape = shd.shape
        return buf, shd

    # return a reference to whatever data was sent
    def getData(self):
        if self['proxy_type'] == ProxyType.np_memmap:
            shd = np.memmap(self['shdf'], dtype=self['dtype'], mode='r', shape=self['shape'])

            # make this look like a normal numpy array, since 
            # functions like np.copy() don't work the same.
            buf = np.frombuffer(shd.data, dtype=shd.dtype)
            buf.shape = shd.shape
            return buf
        elif self['proxy_type'] == ProxyType.np_ndarray:
            return self['data']
        elif self['proxy_type'] == ProxyType.segmented:
            log.error('Segmented Type: this IF requires a list of segment proxy objects')
            return

    # all segments must pass through the proxy separately
    def getDataFromSegments(self, segments):
        if segments[0]['proxy_type'] == ProxyType.segmented:
            if segments[0]['seg_type'] == ProxyType.np_ndarray:
                return self._assembleNDArraySegments(segments)



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


import gc
import time
import numpy as np # for 32bit-Pipe hack
import traceback
import multiprocessing
import platform

import gpi
from gpi import QtCore
from .dataproxy import DataProxy, ProxyType
from .defines import GPI_PROCESS, GPI_THREAD, GPI_APPLOOP
from .logger import manager
from .sysspecs import Specs

# start logger for this module
log = manager.getLogger(__name__)

# Python 3.8 - need to explicitly declare fork for MacOS
if platform.system() == 'Windows':
    multiprocessing_context = multiprocessing.get_context('spawn')
else:
    multiprocessing_context = multiprocessing.get_context('fork')

class ReturnCodes(object):

    # Return codes from the functor have specific meaning to the node internals.
    InitUIError = 2
    ValidateError = 1
    Success = 0
    ComputeError = -1

    def isComputeError(self, ret):
        return ret == self.ComputeError
    def isValidateError(self, ret):
        return ret == self.ValidateError
    def isInitUIError(self, ret):
        return ret == self.InitUIError

    # Return codes from the nodeAPI (i.e. compute(), validate() and initUI())
    # are either success or failure for each function.
    def isSuccess(self, ret):
        return (ret is None) or (ret == 0)
    def isError(self, ret):
        return (ret is not None) and (ret != 0)

Return = ReturnCodes() # make a global copy

def ExecRunnable(runnable):
    tp = QtCore.QThreadPool.globalInstance()
    #print 'active threads: ', tp.activeThreadCount()
    #print 'expiry timeout: ', tp.expiryTimeout()
    #print 'maxThreadCount: ', tp.maxThreadCount()
    tp.start(runnable)

class GPIRunnable(QtCore.QRunnable):
    def __init__(self, func):
        super(GPIRunnable, self).__init__()
        self.run = func
        self.setAutoDelete(True)

class GPIFunctor(QtCore.QObject):
    '''A common parent API for each execution type (i.e. ATask, PTask, TTask).
    Handles the data communications to and from each task type. '''

    finished = gpi.Signal(int)
    terminated = gpi.Signal()
    applyQueuedData_finished = gpi.Signal()
    _setData_finished = gpi.Signal()

    def __init__(self, node, parent=None):
        super(GPIFunctor, self).__init__(parent)
        self._node = node
        self._title = node.name
        self._func = node.getModuleCompute()
        self._validate = node.getModuleValidate()
        self._retcode = None
        self._validate_retcode = None

        # for applying data when a GPI_PROCESS is finished
        # this is done in a thread to keep the GUI responsive
        self._applyData_thread = None
        self._setData_finished.connect(self.applyQueuedData_finalMatter)
        self.applyQueuedData_finished.connect(self.finalMatter)
        self._ap_st_time = 0

        # flag for segmented types that need reconstitution on this side
        self._segmentedDataProxy = False

        # For Windows just make them all apploops for now to be safe
        self._execType = node._nodeIF.execType()
        if Specs.inWindows() and (self._execType == GPI_PROCESS):
        # if (self._execType == GPI_PROCESS):
            log.info("init(): <<< WINDOWS Detected >>> Forcing GPI_PROCESS -> GPI_THREAD")
            self._execType = GPI_THREAD
            # self._execType = GPI_APPLOOP

        self._label = node._nodeIF.getLabel()
        self._isTerminated = False
        self._compute_start = 0

        self._manager = None
        self._proxy = None
        self._proc = None
        if self._execType == GPI_PROCESS:
            log.debug("init(): set as GPI_PROCESS: "+str(self._title))
            self._manager = multiprocessing_context.Manager()
            self._proxy = self._manager.list()
            self._proc = PTask(self._func, self._title, self._label, self._proxy)

            # apply data in a thread to make the GUI more responsive
            self._applyData_thread = GPIRunnable(self.applyQueuedData_setData)

        elif self._execType == GPI_THREAD:
            log.debug("init(): set as GPI_THREAD: "+str(self._title))
            self._proc = TTask(self._func, self._title, self._label, self._proxy)

        else:  # default to GPI_APPLOOP
            log.debug("init(): set as GPI_APPLOOP: "+str(self._title))
            self._proc = ATask(self._func, self._title, self._label, self._proxy)

        self._proc.finished.connect(self.computeFinished)
        # In Qt5, terminated was removed: its emission wasn't gauranteed
        # self._proc.terminated.connect(self.computeTerminated)


    def execType(self):
        return self._execType

    def terminate(self):
        self._isTerminated = True
        self.cleanup()
        self._proc.terminate()
        # self.wait() # so that something is waiting
        self.computeTerminated()

    def cleanup(self):
        # make sure the proxy manager for processes is shutdown.
        if self._manager:
            self._manager.shutdown()

        # try to minimize leftover memory from the segmented array transfers
        # force cleanup of mmap
        #if self._segmentedDataProxy:
        gc.collect()

    def curTime(self):
        return time.time() - self._compute_start

    def start(self):
        self._compute_start = time.time()

        # VALIDATE
        # temporarily trick all widget calls to use GPI_APPLOOP for validate()
        tmp_exec = self._execType
        self._execType = GPI_APPLOOP
        try:
            self._validate_retcode = self._validate()
        except:
            self._validate_retcode = 1 # validate error
        self._execType = tmp_exec

        # None as zero

        # send validate() return code thru same channels
        if self._validate_retcode != 0 and self._validate_retcode is not None:
            self._node.appendWallTime(time.time() - self._compute_start)
            self.finished.emit(1) # validate error
            return

        # COMPUTE
        if self._execType == GPI_PROCESS:
            log.debug("start(): buffer process parms")
            self._node._nodeIF.bufferParmSettings()

            # keep objects on death-row from being copied into processes
            # before they've finally terminated. -otherwise they'll try
            # and terminate within child process and cause a fork error.
            log.debug('start(): garbage collect before spawning GPI_PROCESS')
            gc.collect()
            
        log.debug("start(): call task.start()")
        self._proc.start()

    def wait(self):
        self._proc.wait()

    def isRunning(self):
        return self._proc.isRunning()

    def returnCode(self):
        return self._retcode

    # GPI_PROCESS support
    def addToQueue(self, item):
        # add internal calls (port, widget, retcode...)
        # to a queue that is processed after compute()
        self._proc._proxy.append(item)

    def computeTerminated(self):
        self.terminated.emit()

    def computeFinished(self):
        if self._execType == GPI_PROCESS:
            self.applyQueuedData()

        else:
            self._retcode = 0 # success
            if Return.isError(self._proc._retcode):
                self._retcode = Return.ComputeError
            self.finalMatter()

    def finalMatter(self):
        log.info("computeFinished():Node \'"+str(self._title)+"\': compute time:"+str(time.time() - self._compute_start)+" sec.")
        self._node.appendWallTime(time.time() - self._compute_start)
        self.finished.emit(self._retcode) # success

    def applyQueuedData_setData(self):

        for o in self._proxy:
            try:
                log.debug("applyQueuedData_setData(): apply object "+str(o[0])+', '+str(o[1]))
                if o[0] == 'setData':
                    # DataProxy is used for complex data types like numpy
                    if type(o[2]) is DataProxy:

                        # segmented types must be gathered before reassembly
                        if o[2].isSegmented():
                            log.debug("seg proxy is True")
                            self._segmentedDataProxy = True
                        else:
                            log.debug("o[2].getData()")
                            self._node.setData(o[1], o[2].getData())

                    # all other simple types get set directly
                    else:
                        log.debug("direct setData()")
                        self._node.setData(o[1], o[2])
            except:
                log.error("applyQueuedData() failed. "+str(traceback.format_exc()))
                self._retcode = Return.ComputeError
                self._setData_finished.emit()

        # Assemble Segmented Data
        if self._segmentedDataProxy:
            log.warn("Using segmented data proxy...")
            # group all segmented types
            oportData = [ o for o in self._proxy if (o[0] == 'setData') and (type(o[2]) is DataProxy) ]
            # take only those that are segmented
            oportData = [ o for o in oportData if o[2].isSegmented() ]
            # consolidate all outports with large data
            largeports = set([ o[1] for o in oportData ])

            for port in largeports:
                log.info("applyQueuedData(): ------ APPENDING SEGMENTED PROXY OBJECTS")

                # gather port segs
                curport = [o for o in oportData if o[1] == port]

                # gather all DataProxy segs
                segs = [ o[2] for o in curport ]
                buf = DataProxy().getDataFromSegments(segs)

                # if the pieces fail to assemble move on
                if buf is None:
                    log.warn("applyQueuedData(): segmented proxy object failed to assemble, skipping...")
                    continue

                self._node.setData(port, buf)

        # run self.applyQueuedData_finalMatter()
        self._setData_finished.emit()

    def applyQueuedData(self):
        # Replay all external compute() events after execution.
        # This ensures that the target is not being used by another set method.
        self._ap_st_time = time.time()
        if self._isTerminated:
            return
        log.debug("applyQueuedData(): Sending data to main loop...")
        if len(self._proxy) == 0:
            log.debug("applyQueuedData(): no data in output queue. Terminated.")
            self.computeTerminated()
            return

        self._segmentedDataProxy = False
        for o in self._proxy:
            try:
                log.debug("applyQueuedData(): apply object "+str(o[0])+', '+str(o[1]) )
                if o[0] == 'retcode':
                    self._retcode = o[1]
                    if Return.isError(self._retcode):
                        self._retcode = Return.ComputeError
                    else:
                        self._retcode = 0 # squash Nones
                if o[0] == 'modifyWdg':
                    self._node.modifyWdg(o[1], o[2])
                if o[0] == 'setReQueue':
                    self._node.setReQueue(o[1])
            except:
                log.error("applyQueuedData() failed. "+str(traceback.format_exc()))
                self._retcode = Return.ComputeError

        # transfer all setData() calls to a thread
        log.debug("applyQueuedData(): run _applyData_thread")
        ExecRunnable(self._applyData_thread)

    def applyQueuedData_finalMatter(self):

        if Return.isComputeError(self._retcode):
            self.finished.emit(self._retcode)

        elapsed = (time.time() - self._ap_st_time)
        log.info("applyQueuedData(): time (total queue): "+str(elapsed)+" sec")

        # shutdown the proxy manager
        self.cleanup()

        # start self.finalMatter
        self.applyQueuedData_finished.emit()


class PTask(multiprocessing_context.Process, QtCore.QObject):
    '''A forked process node task. Memmaps are used to communicate data.

    NOTE: The process-type has to be checked periodically to see if its alive,
    from the spawning process.
    '''

    finished = gpi.Signal()
    terminated = gpi.Signal()

    def __init__(self, func, title, label, proxy):
        multiprocessing_context.Process.__init__(self)
        QtCore.QObject.__init__(self)
        self._func = func
        self._title = title
        self._label = label
        self._proxy = proxy
        self._cnt = 0

        # Since we don't know when the process finishes
        # probe at regular intervals.
        # -it would be nicer to have the process check-in with the GPI
        #  main proc when its done.
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self.checkProcess)
        self._timer.start(10)  # 10msec update

    def run(self):
        # This try/except is only good for catching compute() exceptions
        # not run() terminations.
        try:
            self._proxy.append(['retcode', self._func()])
        except:
            log.error('PROCESS: \''+str(self._title)+'\':\''+str(self._label)+'\' compute() failed.\n'+str(traceback.format_exc()))
            self._proxy.append(['retcode', Return.ComputeError])

    def terminate(self):
        self._timer.stop()
        super(PTask, self).terminate()

    def wait(self):
        self.join()

    def isRunning(self):
        return self.is_alive()

    def retcodeExists(self):
        for o in self._proxy:
            if o[0] == 'retcode':
                return True
        return False

    def checkProcess(self):
        if self.is_alive():
            return
        # else if its not alive:
        self._timer.stop()
        if self.retcodeExists():
            # we assume its termination was deliberate.
            self.finished.emit()
        else:
            self.terminated.emit()


class TTask(QtCore.QThread):
    '''A QThread based node runner.  Data is communicated directly.

        NOTE: The thread-type emits a signal when its finished:
            gpi.Signal.finished()
            gpi.Signal.terminated()
    '''

    def __init__(self, func, title, label, proxy):
        super(TTask, self).__init__()
        self._func = func
        self._title = title
        self._label = label
        self._proxy = proxy
        self._retcode = None

        # allow thread to terminate immediately
        # NOTE: doesn't seem to work
        self.setTerminationEnabled(True)

    def terminate(self):
        # Threads don't die as well as processes right now,
        # so just let them run off in the background.
        log.warn("WARNING: Terminated QThread-Node is backgrounded as a zombie.")
        self.exit()  # terminate when finished

    def run(self):
        # This try/except is only good for catching compute() exceptions
        # not run() terminations.
        try:
            self._retcode = self._func()
            log.info("TTask _func() finished")
        except:
            log.error('THREAD: \''+str(self._title)+'\':\''+str(self._label)+'\' compute() failed.\n'+str(traceback.format_exc()))
            self._retcode = Return.ComputeError


class ATask(QtCore.QObject):
    '''App-Loop or Main-Loop executed task.  This will block GUI updates.  Data
    is communicated directly.

        NOTE: The apploop-type blocks until finished, obviating the need for
        signals or timer checks
    '''

    finished = gpi.Signal()
    terminated = gpi.Signal()

    def __init__(self, func, title, label, proxy):
        super(ATask, self).__init__()
        self._func = func
        self._title = title
        self._label = label
        self._proxy = proxy
        self._cnt = 0

    def run(self):
        # This try/except is only good for catching compute() exceptions
        # not run() terminations.
        try:
            self._retcode = self._func()
        except:
            log.error('APPLOOP: \''+str(self._title)+'\':\''+str(self._label)+'\' compute() failed.\n'+str(traceback.format_exc()))
            self._retcode = Return.ComputeError

    def terminate(self):
        pass  # can't happen b/c blocking mainloop

    def wait(self):
        pass  # can't happen b/c blocking mainloop

    def isRunning(self):
        pass  # can't happen b/c blocking mainloop

    def quit(self):
        pass

    def start(self):
        self.run()
        self.finished.emit()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

'''Miscellaneous constants used throughout GPI. These could be for z-buffer,
internal type conventions, uri conventions, etc...
'''


PREFIX='/opt/anaconda1anaconda2anaconda3'

import os
import sys
import inspect
import tempfile

#import gpi
from gpi import QtCore, QtGui, QtWidgets
from .logger import manager


# from:
# http://stackoverflow.com/questions/287871/print-in-terminal-with-colors-using-python
class Cl:
    HDR = '\033[95m'
    OKBL = '\033[94m'
    OKGR = '\033[92m'
    WRN = '\033[93m'
    FAIL = '\033[91m'
    ESC = '\033[0m'

# stringify and color - string-warn
def stw(s):
    return Cl.WRN+str(s)+Cl.ESC


# start logger for this module
log = manager.getLogger(__name__)

def terminalBell():
    #log.dialog('<< BELL >>')
    print('\a')

def GetHumanReadable_bytes(size, precision=2):
    # change size in bytes (int) to a string with nice display units
    suffixes = ['B', 'KB', 'MB', 'GB', 'TB']
    suffixIndex = 0
    while int(size) > 1024:
        suffixIndex += 1
        size = size / 1024.0
    return "%.*f %s" % (precision, size, suffixes[suffixIndex])

def GetHumanReadable_time(t, precision=3):
    # change seconds (float) to a string with nice display units
    t *= 1000.0  # sec -> msec
    suffixes = ['msec', 'sec', 'min', 'hrs', 'days', 'mo', 'yrs']
    div = [1000.0, 60.0, 60.0, 24.0, 30.0, 12.0]
    suffixIndex = 0
    while t > div[suffixIndex]:
        t = t / div[suffixIndex]
        suffixIndex += 1
        if suffixIndex >= len(div):
            break
    return "%.*f %s" % (precision, t, suffixes[suffixIndex])

def ThisFilePath():
    return os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe().f_back)))


# recursion limit info
log.info('Default Recursion Limit: '+str(sys.getrecursionlimit()))
# sys.setrecursionlimit(1500)
log.info('Set Recursion Limit: '+str(sys.getrecursionlimit()))

# location of gpi documents in the packaged distro
GPI_DOCS_DIR = PREFIX+'/share/doc/gpi'

# Node and module paths, local bundle should be searched first.
# The GPI_CWD only gives the path where gpi was invoked, not
# where it is located.
GPI_CWD = os.path.dirname(sys.argv[0])
#sys.path.insert(0, GPI_CWD)

RES_DIR = None
#if os.path.exists(GPI_CWD + "/_res"):
#    RES_DIR = GPI_CWD + "/_res"
#    sys.path.insert(0, RES_DIR)  # for packaging Linux
#elif os.path.exists(GPI_CWD + "/Contents/Resources/_res"):
#    RES_DIR = GPI_CWD + "/Contents/Resources/_res"
#    sys.path.insert(0, RES_DIR)  # for packaging OSX
#else:
#    RES_DIR = GPI_CWD + "/../"
#    sys.path.insert(0, GPI_CWD + "/../")  # for basic kcii setup
#log.debug(sys.path)

GPI_PKG_PATH=os.path.dirname(os.path.abspath( __file__ ))  # get location of THIS gpi python-package
LOGO_PATH = GPI_PKG_PATH+"/graphics/logo.png"
if not os.path.exists(LOGO_PATH):
    log.error("can't find logo.")
ICON_PATH = GPI_PKG_PATH+"/graphics/iclogo.png"
if not os.path.exists(ICON_PATH):
    log.error("can't find icon.")
PLOGO_PATH = GPI_PKG_PATH+"/graphics/slogo.png"
if not os.path.exists(PLOGO_PATH):
    log.error("can't find splash logo.")

# shared memory handles
GPI_SHDM_PATH_PREFIX = 'com.gpilab.GPI'
GPI_SHDM_PATH = tempfile.gettempdir()+'/'+GPI_SHDM_PATH_PREFIX
try:
    os.mkdir(GPI_SHDM_PATH)
    log.info('using shm path: '+GPI_SHDM_PATH)
except:
    log.info(GPI_SHDM_PATH+' already exists')
    if not os.access(GPI_SHDM_PATH, os.R_OK | os.W_OK | os.X_OK):
        GPI_SHDM_PATH = tempfile.mkdtemp(prefix=GPI_SHDM_PATH_PREFIX+'_')
        log.info('using shm path: '+GPI_SHDM_PATH)

#if os.path.exists(GPI_CWD + "/graphics/icons/logo.png"):
#    LOGO_PATH = GPI_CWD + "/graphics/icons/logo.png"
#elif RES_DIR:
#    if os.path.exists(RES_DIR + "/logo.png"):
#        LOGO_PATH = RES_DIR + "/logo.png"
#if LOGO_PATH is None:
#    log.error("can't find logo.")

# qt custom types
UserTYPE = QtWidgets.QGraphicsItem.UserType
EdgeTYPE = UserTYPE + 1
NodeTYPE = UserTYPE + 2
PortTYPE = UserTYPE + 3
InPortTYPE = UserTYPE + 4
OutPortTYPE = UserTYPE + 5
WidgetTYPE = UserTYPE + 6
ExternalNodeType = UserTYPE + 7
ExternalType = UserTYPE + 8
MacroNodeEdgeType = UserTYPE + 9
EdgeNodeType = UserTYPE + 10
PortEdgeType = UserTYPE + 11

def isMacroChildNode(inst):
    '''Determine if its a child node of a macro-node.
    '''
    if hasattr(inst, 'Type'):
        if inst.Type == PortEdgeType:
            return True
    return False

def isNode(inst):
    '''Determine if its a Node or descendent of a node.
    '''
    if hasattr(inst, 'Type'):
        if inst.Type == NodeTYPE:
            return True
    return False

def isWidget(inst):
    '''Determine if its a GPI Widget or descendent of a GenericWidgetGroup.
    '''
    if hasattr(inst, 'GPIWdgType'):
        if inst.GPIWdgType == WidgetTYPE:
            return True
    return False

def isGPIType(inst):
    '''Determine if the object is a GPIType description.
    '''
    if hasattr(inst, 'GPIType'):
        if inst.GPIType == ExternalType:
            return True
    return False

def isExternalNode(inst):
    '''Determine if the object is an ExternalNodeType description (NodeAPI).
    '''
    if hasattr(inst, 'GPIExtNodeType'):
        if inst.GPIExtNodeType == ExternalNodeType:
            return True
    return False


# global definitions
REQUIRED = 100
OPTIONAL = 200

# execution types
GPI_THREAD = 1000
GPI_PROCESS = 2000
GPI_APPLOOP = 3000

# limits
GPI_FLOAT_MAX = sys.float_info.max
GPI_FLOAT_MIN = -sys.float_info.max
GPI_FLOAT_MINSTEP = sys.float_info.min
GPI_INT_MAX = pow(2, 31) - 1  # max QWidgets can use
GPI_INT_MIN = -pow(2, 31)  # min QWidgets can use

# node state events
GPI_WIDGET_EVENT = '_WDG_EVENT_'
GPI_PORT_EVENT = '_PORT_EVENT_'
GPI_INIT_EVENT = '_INIT_EVENT_'
GPI_REQUEUE_EVENT = '_REQUEUE_EVENT_'

# python module file types
GPI_PYMOD_EXTS = ['.py', '.pyc', '.pyo', '.pyd']
GPI_PYMOD_PRE_EXT = '_GPI'
GPI_TYPE_EXTS = ['.py', '.pyc', '.pyo', '.pyd']
GPI_TYPE_PRE_EXT = '_GPITYPE'
GPI_NET_EXTS = ['.net']

def isGPIModFile(path):
    '''Determine if the path exists, isfile, endswith _GPI, and valid ext.
    '''
    if os.path.isfile(path):
        bname, ext = os.path.splitext(path)
        if bname.endswith(GPI_PYMOD_PRE_EXT) and (ext.lower() in GPI_PYMOD_EXTS):
            return True
    return False

def isGPITypeFile(path):
    '''Determine if the path exists, isfile, endswith _GPITYPE, and valid ext.
    '''
    if os.path.isfile(path):
        bname, ext = os.path.splitext(path)
        if bname.endswith(GPI_TYPE_PRE_EXT) and (ext.lower() in GPI_PYMOD_EXTS):
            return True
    return False

def isGPINetworkFile(path):
    '''Determine if the path exists, isfile, and valid ext.
    '''
    if os.path.isfile(path):
        bname, ext = os.path.splitext(path)
        if ext.lower() in GPI_NET_EXTS:
            return True
    return False


def TranslateFileURI(uri):
    '''Translate the given uri to a local path.
    '''
    # Check for cwd:// protocol, this will
    # grab cwd and replace it with the GPI_CWD.
    # NOTE: useful for packaging example networks
    # that point to relative data paths.
    if uri.startswith('cwd://'):
        # uri = GPI_CWD+'/'+uri.split('cwd://')[1]
        uri = RES_DIR + '/' + uri.split('cwd://')[1]
    # just strip the 'file://' URI
    elif uri.startswith('file://'):
        uri = uri.split('file://')[1]
    # exapand user $HOME
    elif uri.startswith('~/'):
        uri = os.path.expanduser(uri)

    return uri


def getKeyboardModifiers(supress=False):
    '''Just for testing purposes, replaces
    QtWidgets.QApplication.keyboardModifiers()
    '''
    #supress = True  # force off
    mod = QtWidgets.QApplication.keyboardModifiers()
    if not supress:
        if mod == QtCore.Qt.NoModifier:
            log.debug(" KeyboardMods: NoModifier")
        if mod == QtCore.Qt.ShiftModifier:
            log.debug(" KeyboardMods: ShiftModifier")
        if mod == QtCore.Qt.ControlModifier:
            log.debug(" KeyboardMods: ControlModifier")
        if mod == QtCore.Qt.AltModifier:
            log.debug(" KeyboardMods: AltModifier")
        if mod == QtCore.Qt.MetaModifier:
            log.debug(" KeyboardMods: MetaModifier")
        if mod == QtCore.Qt.KeypadModifier:
            log.debug(" KeyboardMods: KeypadModifier")
        if mod == QtCore.Qt.GroupSwitchModifier:
            log.debug(" KeyboardMods: GroupSwitchModifier")
    return mod


def printMouseEvent(event, supress=False):
    #supress = True
    if not supress:
        if event.buttons() & QtCore.Qt.LeftButton:
            log.debug(" Mouse: LeftButton")
        if event.buttons() & QtCore.Qt.RightButton:
            log.debug(" Mouse: RightButton")
        if event.buttons() & QtCore.Qt.MidButton:
            log.debug(" Mouse: MidButton")




#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# All methods, defs and members needed by node developers must be imported
# automatically into the gpi namespace.  This way users only need to
# import gpi, then use things like gpi.REQUIRED for ports etc...
import warnings
warnings.filterwarnings("ignore", ".*Applications.GPI.*import.*")

import os
import time

GPI_PKG_PATH=os.path.dirname(os.path.abspath( __file__ ))  # get location of THIS gpi python-package
VERSION_FPATH=GPI_PKG_PATH+'/VERSION'
VERSION = '1.4.4'
__version__ = VERSION
RELEASE_DATE = '2022-03-31'
try:
    with open(VERSION_FPATH, 'r') as f:
        for l in f.readlines():
            if l.count('PKG_VERSION'):
                VERSION = l.split(':')[-1].strip()
            if l.count('BUILD_DATE'):
                RELEASE_DATE = l.split(':')[-1].strip()
except:
    pass

# Print version info each time.
_version = 'GPI '+VERSION+' ('+RELEASE_DATE+')'
_copyright = ''
_disclaimer = '''\
This program comes with ABSOLUTELY NO WARRANTY; see the LICENSE for details.

    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE SOFTWARE
MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC PURPOSES.  YOU
ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR USE IN ANY HIGH
RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED TO LIFE SUPPORT
OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO WARRANTY AND HAS
NO LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY HIGH RISK OR STRICT
LIABILITY ACTIVITIES.
'''

print((_version+'  '+_copyright+'\n'+_disclaimer))

# transitioning tool
from . import qtapi
QtCore = qtapi.QtCore
QtGui = qtapi.import_module("QtGui")
QtWidgets = qtapi.import_module("QtWidgets")
QtMultimedia = qtapi.import_module("QtMultimedia")
QT_API_NAME = qtapi.API_NAME
QtWebKit = qtapi.QtWebKit
QtWebKitWidgets = qtapi.QtWebKitWidgets
QtWebEngineWidgets = qtapi.QtWebEngineWidgets
QWebView = qtapi.QWebView
QtOpenGL = qtapi.QtOpenGL
Qimport = qtapi.import_module
Signal = qtapi.Signal
Slot = qtapi.Slot
Property = qtapi.Property

# redirect stdout
#from . import console

# commandline parsing
from . import cmd

# all global vars
from .defines import *

# numba decorators
from .numba_stub import autojit, jit

# default type class
from .defaultTypes import *

# all widget elements and default widgets
from .widgets import *

# widget menu
from .nodeAPI import *

from .remote import *
remote = run_on_server

from .parallel import *
parallel = Parallel()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

''' A pymod loader for managing loaded nodes, types, and widgets. '''


import os
import imp
import sys
import traceback
import py_compile

# gpi
from .defines import GPI_PYMOD_EXTS
from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)

def consolidatePaths(plist):
    '''Take the set() of all abspaths().  Doesn't guarantee original order.
    '''
    nlist = []
    for l in plist:
        nlist.append(os.path.abspath(l))
    nlist = list(set(nlist))
    return nlist

def appendSysPath(path):
    '''Put a new path at the end of sys.path if it doesn't already exist.
    '''
    if path not in sys.path:
        #print 'adding path to sys: '+path
        sys.path.append(path)

def consolidateSysPath():
    '''Run consolitatePaths() on sys.path.
    '''
    sys.path = consolidatePaths(sys.path)

def PKGroot(fullpath):
    '''See if the file or directory pointed to by fullpath is a package dir
    or subpackage dir.  Return the highest package dir.
    '''
    # strip off filename if included
    path = fullpath
    if not os.path.isdir(fullpath):
        path = os.path.dirname(fullpath)

    # skip ./GPI dir
    if os.path.basename(path) == 'GPI':
        path = os.path.dirname(path)

    # find root package dir
    root_found = False
    while not root_found:

        # check if cur dir is part of a package
        if os.path.isfile(path + '/__init__.py'):
            path = os.path.dirname(path)
        else:
            root_found = True

    return path


def loadMod(fullpath):
    '''Load modules .py or .pyc from the given path and store in sys.modules
    using the fullpath as the key.  This will allow all plugins and node
    descriptions to be unique, even if they have the same name.
    '''

    if not os.path.isfile(fullpath):
        log.error('The supplied path is not a file: '+str(fullpath))
        return None

    # exclude the file extension to allow reloads
    store_name, ext = os.path.splitext(fullpath)
    if ext not in GPI_PYMOD_EXTS:
        log.error('The filename is not a valid pymod: '+str(fullpath))
        return None

    # make import params
    fp = open(fullpath, "rb")
    description = (ext, 'rb', 1)

    # load .py
    if ext == '.py':

        # only attempt compilation if the directory is writeable 
        #   -this helps with distributed libraries.
        if os.access(os.path.dirname(fullpath), os.W_OK):

            # Force compile every time b/c some virtual machines somehow get
            # incorrect timestaps which causes node updates not to be taken.
            try:
                py_compile.compile(fullpath, doraise=True)
                log.info('SUCCESS: '+fullpath)
            except:
                log.error(str(traceback.format_exc()) + '\nFAILED:'+fullpath)

        else:
            log.info('Cannot compile, permission denied: '+str(fullpath))

        try:
            mod = imp.load_module(store_name, fp, fullpath, description)
        except: # ImportError:
            log.error(str(fullpath)+' module failed to load in loadMod(.py) with:\n' + str(traceback.format_exc()))
            return None
        finally:
            if fp:
                fp.close()

    # load compiled .pyc, etc...
    else:
        try:
            mod = imp.load_compiled(store_name, fullpath, fp)
        except: # ImportError:
            log.error(str(fullpath)+' module failed to load in loadMod(\'compiled\') with:\n' + str(traceback.format_exc()))
            return None
        finally:
            if fp:
                fp.close()

    return mod


def findAndLoadMod(name, path=None, store_name=None):
    '''Load modules .py or .pyc from sys.path or path, if given.
    'path' must be a list.

    DEPRECATED
        -not sure when this stopped being used or if find_module()
        affords us anything over the loadMod().
    '''

    if store_name is None:
        store_name = name

    if path is not None:
        cnt = 0
        for p in path:
            if not os.path.isdir(p):
                log.error('The supplied path is not a directory: '+str(p))
                cnt += 1
            if len(path) == cnt:
                log.error('None of the supplied paths are valid dirs, skipping load().')
                return None

    # make import params
    try:
        if path is not None:
            fp, pathname, description = imp.find_module(name, path)
        else:
            fp, pathname, description = imp.find_module(name)
    except: # ImportError:
        log.error('Failed to locate module: '+str(name))
        return None

    # load .py
    if description[0] == '.py':

        # Force compile every time b/c some virtual machines somehow get
        # incorrect timestaps which causes node updates not to be taken.
        #try:
        #    compileall.compile_file(pathname)
        #    print pathname
        #    log.dialog('findAndLoadMod: py compiled.')
        #except:
        #    log.dialog('findAndLoadMod: py not compiled.')

        try:
            mod = imp.load_module(store_name, fp, pathname, description)
        except: # ImportError:
            log.error(str(name)+' module failed to load in findAndLoadMod(.py) with:\n' + str(traceback.format_exc()))
            return None
        finally:
            if fp:
                fp.close()

    # load compiled .pyc, etc...
    else:
        try:
            mod = imp.load_compiled(store_name, pathname, fp)
        except: # ImportError:
            log.error(str(name)+' module failed to load in findAndLoadMod(compiled) with:\n' + str(traceback.format_exc()))
            return None
        finally:
            if fp:
                fp.close()

    return mod



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

'''
File-node associations for data drag'n drop.
'''

import os
import sys

# gpi
from .catalog import Catalog, CatalogObj
from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)

# bind<num> = ext, key or name, widget
# NOTE: currently this must be in alphabetical order to be correct in the
# .gpirc or --config display.
bind_1 = ('.csv', 'gpi_core.fileIO.ReadCSV', 'File Browser')
bind_2 = ('.data', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_3 = ('.hdf5', 'gpi_core.fileIO.ReadHDF5', 'File Browser')
bind_4 = ('.jpg', 'gpi_core.fileIO.ReadImage', 'File Browser')
bind_5 = ('.lab', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_6 = ('.list', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_7 = ('.mat', 'gpi_core.fileIO.ReadMatlab', 'File Browser')
bind_8 = ('.npy', 'gpi_core.fileIO.ReadNPY', 'File Browser')
bind_9 = ('.par', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_10 = ('.pickle', 'gpi_core.fileIO.ReadPickled', 'File Browser')
bind_11 = ('.png', 'gpi_core.fileIO.ReadImage', 'File Browser')
bind_12 = ('.raw', 'gpi_core.fileIO.ReadRaw', 'File Browser')
bind_13 = ('.rec', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_14 = ('.sin', 'gpi_core.fileIO.ReadPhilips', 'File Browser')
bind_15 = ('.xml', 'gpi_core.fileIO.ReadPhilips', 'File Browser')


# make a catalog of associations
Bindings = Catalog()

def isGPIAssociatedFile(fullpath):
    '''Determine if the path exists, isfile, and valid ext.
    '''
    if os.path.isfile(fullpath):
        bpath, ext = os.path.splitext(fullpath)
        if str(ext).lower() in list(Bindings.keys()):
            return True
    return False

def isGPIAssociatedExt(ext):
    '''Determine if the path exists, isfile, and valid ext.
    '''
    if str(ext).lower() in list(Bindings.keys()):
        return True
    return False

class BindCatalogItem(CatalogObj):
    '''Holds all necessary things for file associations.
    '''

    def __init__(self, bind):
        self.ext = bind[0].lower() # all extensions should be case-insensitive
        self.node = bind[1]
        self.wdg = bind[2]

    def key(self):
        # its most likely that this will be looked up by extension
        return self.ext.lower()

    def merge(self, e):
        # allow defaults to be overwritten
        log.info('\'' + str(self.key()) + '\' extension for \'' +str(e.node) + '\' is already assigned to \'' + str(self.node) + '\', replacing.')
        self.ext = e.ext
        self.node = e.node
        self.wdg = e.wdg

    def asTuple(self):
        return (self.ext, self.node, self.wdg)

    def __str__(self):
        return str(self.ext) + ' ' + str(self.node) + ' ' + str(self.wdg)

# load the default bindings
for b in [ x for x in dir(sys.modules[__name__]) if x.startswith('bind_') ]:
    if hasattr(sys.modules[__name__], b):
        item = BindCatalogItem(getattr(sys.modules[__name__], b))
        Bindings.append(item)

#print Bindings
